use std::error::Error;

use crate::mode::*;

/// [`Atomic`] equivalent of [`ParseError`].
///
/// [`ParseError`]: crate::rcore::ParseError
pub type AParseError<A> = <A as AtomicBase>::AParseError;

pub type AParseResult<A> = Result<A, AParseError<A>>;

/// [`Atomic`] base.
pub trait AtomicBase: 'static + Send + Sync + Send + Clone + Serializable {
    /// Equivalent of [`FactoryBase::ParseError`].
    ///
    /// [`FactoryBase::ParseError`]: crate::rcore::FactoryBase::ParseError
    type AParseError: Error;
}

/// This trait combines functionality of [`Mentionable`] and [`Factory`],
/// while limiting [`MentionableTop::points_typed`] (and corresponding [`MentionableTop::topology`])
/// to an empty sequence.
///
/// [`Mentionable`]: crate::rcore::Mentionable
/// [`Factory`]: crate::rcore::Factory
/// [`MentionableTop::points_typed`]: crate::rcore::MentionableTop::points_typed
/// [`MentionableTop::topology`]: crate::rcore::MentionableTop::topology
pub trait Atomic: AtomicBase + ParseMode {
    /// Static equivalent of [`FactoryParse::deserialize`].
    ///
    /// [`FactoryParse::deserialize`]: crate::rcore::FactoryParse::deserialize
    fn a_deserialize(stream: impl Stream) -> AParseResult<Self>;

    /// Static equivalent of [`FactoryParse::extend`].
    ///
    /// [`FactoryParse::extend`]: crate::rcore::FactoryParse::extend
    fn a_extend(self, tail: &[u8]) -> AParseResult<Self>;
}
