use super::*;

impl<'a, Ctx: Context<'a>, F: FactoryModeParse<'a, Ctx>> FactoryParse<'a, Ctx> for F {
    fn deserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self> {
        self.mdeserialize(inctx).map(Self::seal)
    }

    fn extend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self> {
        Self::xseal(self.mextend(Self::prepare(mentionable), tail))
    }
}

impl<'a, Ctx: Context<'a>, F: FactoryParse<'a, Ctx> + ParseMode> Factory<'a, Ctx> for F
where
    F::Mtbl: MentionableTop<'a, Ctx>,
{
    type _Mtbl = Self::Mtbl;
}

/// [`FactoryParse`] equivalent of [`ModeResult`].
pub type ModeResultM<'a, Ctx, F, I> = ModeResultP<F, Mtbl<'a, Ctx, F>, ParseError<'a, Ctx, F>, I>;

/// [`FactoryParse`] equivalent of [`ExtensionResult`].
pub type ExtensionResultM<'a, Ctx, F> =
    ExtensionResultP<F, Mtbl<'a, Ctx, F>, ParseError<'a, Ctx, F>>;

/// [`FactoryParse`] equivalent of [`ExtensionSource`].
pub type ExtensionSourceM<'a, Ctx, F> = ExtensionSourceP<F, Mtbl<'a, Ctx, F>>;

/// A more generic version of [`FactoryParse`].
pub trait FactoryModeParse<'a, Ctx: Context<'a>>: FactoryBase<'a, Ctx> + ParseMode {
    /// A more generic version of [`FactoryParse::deserialize`].
    fn mdeserialize<I: InCtx<'a, Ctx>>(&self, inctx: I) -> ModeResultM<'a, Ctx, Self, I>;

    /// A more generic version of [`FactoryParse::extend`].
    fn mextend(
        &self,
        mentionable: ExtensionSourceM<'a, Ctx, Self>,
        tail: &[u8],
    ) -> ExtensionResultM<'a, Ctx, Self>;
}

/// External implementation of [`FactoryModeParse`].
pub trait FactoryModeProxy<'a, Ctx: Context<'a>> {
    /// Associated [`FactoryModeParse`].
    type F: FactoryBase<'a, Ctx> + ParseMode;

    /// External implementation of [`FactoryModeParse::mdeserialize`].
    fn pmdeserialize<I: InCtx<'a, Ctx>>(f: &Self::F, inctx: I) -> ModeResultM<'a, Ctx, Self::F, I>;

    /// External implementation of [`FactoryModeParse::mextend`].
    fn pmextend(
        f: &Self::F,
        mentionable: ExtensionSourceM<'a, Ctx, Self::F>,
        tail: &[u8],
    ) -> ExtensionResultM<'a, Ctx, Self::F>;
}

impl<'a, Ctx: Context<'a>, F: FactoryBase<'a, Ctx> + WithParseMode> FactoryModeParse<'a, Ctx> for F
where
    F::WithMode: FactoryModeProxy<'a, Ctx, F = F>,
{
    fn mdeserialize<I: InCtx<'a, Ctx>>(&self, inctx: I) -> ModeResultM<'a, Ctx, Self, I> {
        <F::WithMode as FactoryModeProxy<'a, Ctx>>::pmdeserialize(self, inctx)
    }

    fn mextend(
        &self,
        mentionable: ExtensionSourceM<'a, Ctx, Self>,
        tail: &[u8],
    ) -> ExtensionResultM<'a, Ctx, Self> {
        <F::WithMode as FactoryModeProxy<'a, Ctx>>::pmextend(self, mentionable, tail)
    }
}

/// For auto-deriving [`RegularFactory`] from concrete implementations.
pub trait CRegularFactory<'a, Ctx: Context<'a>>:
    FactoryBase<'a, Ctx> + ImplMode<Mode = RegularMode>
{
    /// Concrete implementation of [`RegularFactory::rdeserialize`].
    fn crdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self>;
    /// Concrete implementation of [`RegularFactory::rextend`].
    fn crextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self>;
}

/// Mostly same as [`FactoryModeParse`] but requires [`Mode`] to be [`RegularMode`].
pub trait RegularFactory<'a, Ctx: Context<'a>>:
    FactoryBase<'a, Ctx> + ParseMode<Mode = RegularMode>
{
    /// Same as [`FactoryModeParse::mdeserialize`].
    fn rdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self>;
    /// Same as [`FactoryModeParse::mextend`].
    fn rextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self>;
}

impl<'a, Ctx: Context<'a>, F: FactoryModeParse<'a, Ctx> + ParseMode<Mode = RegularMode>>
    RegularFactory<'a, Ctx> for F
{
    fn rdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self> {
        self.mdeserialize(inctx)
    }

    fn rextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self> {
        self.mextend(mentionable, tail)
    }
}

impl<'a, Ctx: Context<'a>, F: CRegularFactory<'a, Ctx>> FactoryModeProxy<'a, Ctx>
    for WithMode<F, RegularMode>
{
    type F = F;

    fn pmdeserialize<I: InCtx<'a, Ctx>>(f: &Self::F, inctx: I) -> ModeResultM<'a, Ctx, F, I> {
        f.crdeserialize(inctx)
    }

    fn pmextend(f: &F, mentionable: Mtbl<'a, Ctx, F>, tail: &[u8]) -> ExtensionResultM<'a, Ctx, F> {
        f.crextend(mentionable, tail)
    }
}
